""" 
cricketstats is a script for getting team and player statistics from the cricsheet.org database for data analysis.
Copyright (C) 2021  Saranga Sudarshan

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>. 
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# If you installed cricketstats by pip import the module with line below
import cricketstats as cks

# If you cloned the github repo uncomment the block below.
"""
import os
import sys
module_path = os.path.abspath(os.path.join("./cricketstats/src/"))
if module_path not in sys.path:
    sys.path.append(module_path)
"""


""" 1. Create a simulation object """
sim = cks.mpmatchsim.matchsim(simteams=["Sri Lanka", "Australia"])

""" 2. Apply the "sim()" method to the simulation object with the necessary arguments. """
# Arguments are mostly lists, with items separated by commas.

# First we can setup our arguments.
# Required input arguments:

statsdatabase="/home/saranga/Downloads/all_json.zip" # Path of cricsheet.org's all matches json zip file on your system. You can download the file at https://cricsheet.org/downloads/
statsfrom_date=(2021,7,1) # The (Year, Month, Day) from which to start search for probability values
statsto_date=(2022,12,31) # The (Year, Month, Day) from which to end search for probability values
statssex="male" # The male or female team stats to search for
statsmatchtype = "T20" # match type to simulated and search stats for Options: ["Test","ODI", "ODM", "T20",] Explanation of what these mean is found at https://cricsheet.org/downloads/
simulations=10 # Number of matches to simulate. Make sure this is divisble by the number of cores you have in your computer otherwise the module will fail.

# Optional Arguments:
inningsorder = None # This sets the order in which teams bat. Eg. ["Sri Lanka", "Australia","Sri Lanka", "Australia"] for a test match # This overrides the toss rng and sets a specific innings order.
rain=False # This sets whether there might be rain affected matches in the simulated games. This is set at constant 90/10% in favour of no rain.

matchscore = None # This sets a score or match situation from which a matches will simulated eg. matchscore = {"Innings 1":["Sri Lanka",10, 300, 90],"Innings 2":["Australia"",10, 300, 90],"Innings 3":["Sri Lanka",10, 300, 90],"Innings 4":["Australia"",0, 1, 1]}. when using this argument, inningsorder must be given to ensure the correct inningsorder is preserved

# Apply sim() method on the simulation object.
sim.sim(statsdatabase=statsdatabase, statsfrom_date=statsfrom_date,statsto_date=statsto_date, statssex=statssex,statsmatchtype=statsmatchtype,simulations=simulations, inningsorder=inningsorder, rain=rain)


# You can use the above template if you want or put all the values inside the brackets of the "sim()" method like I've done below for the object.
# sim1.sim(statsdatabase="/home/saranga/Downloads/all_json.zip", statsfrom_date=(2021,7,1),statsto_date=(2022,12,31), statssex="male",statsmatchtype="Test",simulations=10)


""" 3. Print result. Output is a pandas dataframe. """
# The output of the search is a pandas dataframe.

print(sim.results)

# You can then use all the familiar pandas methods for statistical analysis and plotting.
# Eg. sim.simresults["Winner"].value_counts(normalize=True) # This command will given you winnings percentage for each team.

# If you want to save the sim result as a csv file that you can open in excel you can use the line below.
# sim1.results.to_csv(./YOUR FILE LOCATION)



""" 
cricketstats is a script for getting team and player statistics from the cricsheet.org database for data analysis.
Copyright (C) 2021  Saranga Sudarshan

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>. 
"""
import os
import json

import pandas as pd
import math
import requests
from bs4 import BeautifulSoup

# def playersearch(player):
#     currentdir = os.path.dirname(os.path.abspath(__file__))
#     playerindexfile = open(f"{currentdir}/playerindex.json")
#     playerindex = json.load(playerindexfile)
#     result = {"Player":[],"Batting":[],"Bowling":[]}
#     for eachlist in playerindex["Batting"]:
#         for eachitem in playerindex["Batting"][eachlist]:
#             if player in eachitem and eachitem not in result["Player"]:
#                 result["Player"].append(eachitem)
#                 result["Batting"].append(eachlist)
#     playerfound=False
#     for eachplayer in result["Player"]:
#         result["Bowling"].append(None)
#         for eachlist in playerindex["Bowling"]:
#         #for eachitem in playerindex["Bowling"][eachlist]:
#             if eachplayer in playerindex["Bowling"][eachlist]:
#                 result["Bowling"][result["Player"].index(eachplayer)]=eachlist
#     print(len(result["Player"]))
#     print(len(result["Batting"]))
#     print(len(result["Bowling"]))
#     df = pd.DataFrame(result)
#     return df

def index(playerdatabase,playerindexfile=None):
    if playerindexfile==None:
        currentdir = os.path.dirname(os.path.abspath(__file__))
        indexfile = open(f"{currentdir}/playerindex.json")
        playerindex = json.load(indexfile)
        playerlist = pd.read_csv(playerdatabase)
        newplayers = {}
        #{"Batting":None,"Bowling":None,"Umpire":"No"}
    else:
        playerlist = pd.read_csv(playerdatabase)
        indexfile = open(playerindexfile)
        playerindex = json.load(indexfile)
        newplayers = {}
        #"Player":[],"Batting":[],"Bowling":[],"Umpire":"No"}
    try:
        for eachplayer in playerlist["name"]:
            if eachplayer in playerindex.keys() and playerindex[eachplayer]["Batting"]!=None and playerindex[eachplayer]["Bowling"]!=None:
                continue

            #playerdata = playerlist.loc[playerlist["name"]==f"{eachplayer}", ["key_cricinfo"]]
            if pd.isna(playerlist.at[playerlist.loc[playerlist["name"]==f"{eachplayer}"].index[0], "key_cricinfo"]):
                continue
            playerid = math.floor(playerlist.at[playerlist.loc[playerlist["name"]==f"{eachplayer}"].index[0], "key_cricinfo"])
            print(eachplayer + ", "+ f"https://www.espncricinfo.com/*/content/player/{playerid}.html")
            webpage = requests.get(f"https://www.espncricinfo.com/*/content/player/{playerid}.html")
            playerpage = BeautifulSoup(webpage.content, "html.parser")
            main = playerpage.find(id="main-container")
            if main is None:
                continue
            playerinfo = main.find_all("h5", class_="player-card-description gray-900")
            for element in playerinfo:
                if eachplayer not in playerindex.keys() and ("Right hand bat" in element.text or "Left hand bat" in element.text or "Umpire" in element.text):
                    playerindex[eachplayer]={"Batting":None,"Bowling":None,"Umpire":None}

                # Batting
                if "Right hand bat" in element.text and playerindex[eachplayer]["Batting"]==None:
                    playerindex[eachplayer]["Batting"]="Right hand bat"

                if "Left hand bat" in element.text and playerindex[eachplayer]["Batting"]==None:
                    playerindex[eachplayer]["Batting"]="Left hand bat"
                
                # Bowling
                if ("Right arm fast" in element.text or "Right arm fast medium" in element.text  or "Right arm medium" in element.text) and playerindex[eachplayer]["Bowling"]==None:
                    playerindex[eachplayer]["Bowling"]="Right arm pace"

                if ("Left arm fast" in element.text or "Left arm fast medium" in element.text  or "Left arm medium" in element.text) and playerindex[eachplayer]["Bowling"]==None:
                    playerindex[eachplayer]["Bowling"]="Left arm pace"

                if "offbreak" in element.text and playerindex[eachplayer]["Bowling"]==None:
                    playerindex[eachplayer]["Bowling"]="Right arm off break"
                if "Legbreak" in element.text and playerindex[eachplayer]["Bowling"]==None:
                    playerindex[eachplayer]["Bowling"]="Right arm leg break"
                    
                if "Slow left arm orthodox" in element.text and playerindex[eachplayer]["Bowling"]==None:
                    playerindex[eachplayer]["Bowling"]="Left arm orthodox"

                if ("Left arm wrist spin" in element.text or "Left-arm googly" in element.text) and playerindex[eachplayer]["Bowling"]==None:
                    playerindex[eachplayer]["Bowling"]="Left arm wrist spin"
                # Umpire
                if "Umpire" in element.text and (playerindex[eachplayer]["Umpire"]==None or playerindex[eachplayer]["Umpire"]=="No"):
                    playerindex[eachplayer]["Umpire"]="Yes"
    finally:
        indexfile.close()
        if playerindexfile==None:
            file = open(f"{currentdir}/playerindex.json", "w")
            file.write(json.dumps(playerindex))
            file.close()
        if playerindexfile!=None:
            file = open(playerindexfile, "w")
            file.write(json.dumps(playerindex))
            file.close()
    



""" 
cricketstats is a script for getting team and player statistics from the cricsheet.org database for data analysis.
Copyright (C) 2021  Saranga Sudarshan

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>. 
"""

import os
from timeit import timeit
import pandas as pd
import numpy as np
import multiprocessing as mp
import time
from cricketstats import cricketstats


# TODO record player's scores and balls faced
# TODO fix bowler's runs
# TODO add bowling order, or some mechanism to pick specific bowlers, yeah



class playersim:
    def __init__(self,simteams=None) -> None:
        self.simteams = simteams
        self.simresults = None
        self.simteamstats = None

    def simresultssetup(self, statsmatchtype):
        limitedovers = {"Innings 1 Team":[], "Innings 1 Wickets":[],"Innings 1 Score":[], "Innings 1 Overs":[], "Innings 2 Team":[],"Innings 2 Wickets":[], "Innings 2 Score":[], "Innings 2 Overs":[],"Winner":[]}
        testmatch={"Innings 1 Team":[],"Innings 1 Wickets":[], "Innings 1 Score":[], "Innings 1 Overs":[],"Innings 2 Team":[],"Innings 2 Wickets":[], "Innings 2 Score":[],"Innings 2 Overs":[],"Innings 3 Team":[],"Innings 3 Wickets":[], "Innings 3 Score":[],"Innings 3 Overs":[],"Innings 4 Team":[], "Innings 4 Wickets":[],"Innings 4 Score":[],"Innings 4 Overs":[],"Winner":[]}
        matchtypes={"T20": limitedovers, "ODI": limitedovers,"ODM": limitedovers,"Test": testmatch}
        self.simresults = matchtypes[statsmatchtype]

    def pvaluesearch(self, statsdatabase, statsfrom_date, statsto_date, statssex, statsmatchtype):
        # search stats for p-values
        simplayers=[]
        for eachteam in self.simteams:
            for eachlist in self.simteams[eachteam]:
                simplayers.extend(self.simteams[eachteam][eachlist])

        searchmatchtypes={"T20": ["T20"], "ODI": ["ODI","ODM"],"ODM": ["ODI","ODM"],"Test": ["Test","MDM"],"MDM": ["Test","MDM"]} 

        self.simteamstats = cricketstats.search(players=simplayers)
        self.simteamstats.stats(database=statsdatabase, from_date=statsfrom_date, to_date=statsto_date, matchtype=searchmatchtypes[statsmatchtype], betweenovers=[], innings=[], sex=[statssex], playersteams=[], oppositionbatters=[], oppositionbowlers=[], oppositionteams=[], venue=[], event=[], matchresult=[], superover=None, battingposition=[], bowlingposition=[], fielders=[], sumstats=False)

    def playerP(self,nthinnings,thisinnings, bowlingteam,simteams,simteamstats,thisover,hometeam):

        if hometeam==thisinnings:
            batadv=0.55
            bowladv=1-batadv
        if hometeam==bowlingteam:
            bowladv=0.55
            batadv=1-bowladv
        if hometeam==None:
            batadv=0.5
            bowladv=0.5

        # wicket probabilites
        # wicketfallP = {self.batters[0]:
        # (simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Batter"]==self.batters[0])&(simteamstats.ballresult["Innings Type"]=="Batting")&(simteamstats.ballresult["Balls Faced"]<(self.battersstats[self.batters[0]]+6))&(simteamstats.ballresult["Innings"]==(nthinnings+1))].value_counts(normalize=True,sort=False)*batadv).add(
        #     simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))&(simteamstats.ballresult["Innings"]==(nthinnings+1))].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
        # ,
        # self.batters[1]: 
        # (simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Batter"]==self.batters[1])&(simteamstats.ballresult["Innings Type"]=="Batting")&(simteamstats.ballresult["Balls Faced"]<(self.battersstats[self.batters[0]]+6))&(simteamstats.ballresult["Innings"]==(nthinnings+1))].value_counts(normalize=True,sort=False)*batadv).add(
        #     simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))&(simteamstats.ballresult["Innings"]==(nthinnings+1))].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
        # }

        # # wicket probability all innings fallback
        # if len(wicketfallP[self.batters[0]])<2 or sum(wicketfallP[self.batters[0]])<0.99:
        wicketfallP = {self.batters[0]:
        (simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Batter"]==self.batters[0])&(simteamstats.ballresult["Innings Type"]=="Batting")&(simteamstats.ballresult["Balls Faced"]<(self.battersstats[self.batters[0]]+6))].value_counts(normalize=True,sort=False)*batadv).add(
            simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
        ,
        self.batters[1]: 
        (simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Batter"]==self.batters[1])&(simteamstats.ballresult["Innings Type"]=="Batting")&(simteamstats.ballresult["Balls Faced"]<(self.battersstats[self.batters[0]]+6))].value_counts(normalize=True,sort=False)*batadv).add(
            simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
        }

        # Score probabilites
        # scoreP= {self.batters[0]:
        # (simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Batter"]==self.batters[0])&(simteamstats.ballresult["Balls Faced"]<(self.battersstats[self.batters[0]]+6))&(simteamstats.ballresult["Innings"]==(nthinnings+1))].value_counts(normalize=True,sort=False)*batadv).add(
        #     simteamstats.ballresult['Batter Score'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))&(simteamstats.ballresult["Innings"]==(nthinnings+1))].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
        # ,
        # self.batters[1]:
        # (simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Batter"]==self.batters[1])&(simteamstats.ballresult["Balls Faced"]<(self.battersstats[self.batters[0]]+6))&(simteamstats.ballresult["Innings"]==(nthinnings+1))].value_counts(normalize=True,sort=False)*batadv).add(
        #     simteamstats.ballresult['Batter Score'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))&(simteamstats.ballresult["Innings"]==(nthinnings+1))].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
        # }
        # if len(scoreP[self.batters[0]])==0 or sum(scoreP[self.batters[0]])<0.99:
            #print(f"Innings {nthinnings}, {self.batters[0]}")
        scoreP= {self.batters[0]:
        (simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Batter"]==self.batters[0])&(simteamstats.ballresult["Balls Faced"]<(self.battersstats[self.batters[0]]+6))].value_counts(normalize=True,sort=False)*batadv).add(
            simteamstats.ballresult['Batter Score'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
        ,
        self.batters[1]:
        (simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Batter"]==self.batters[1])&(simteamstats.ballresult["Balls Faced"]<(self.battersstats[self.batters[0]]+6))].value_counts(normalize=True,sort=False)*batadv).add(
            simteamstats.ballresult['Batter Score'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
        }

        

        return wicketfallP,scoreP

    def extras(self,nthinnings,thisinnings, bowlingteam,simteams,simteamstats,thisover):

        extrasP= simteamstats.ballresult["Bowler Extras"].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Ball"]>(thisover))&(simteamstats.ballresult["Innings Ball"]<(thisover+1))].value_counts(normalize=True,sort=False)
        
        fieldingextrasP= simteamstats.ballresult["Fielding Extras"].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))].value_counts(normalize=True,sort=False)

        return extrasP,fieldingextrasP

    def wicketfallbackplayersP(self,nthinnings,thisinnings, bowlingteam,simteams,simteamstats,thisover,hometeam):

        if hometeam==thisinnings:
            batadv=0.55
            bowladv=1-batadv
        if hometeam==bowlingteam:
            bowladv=0.55
            batadv=1-bowladv
        if hometeam==None:
            batadv=0.5
            bowladv=0.5

        inningswicketfallP = {
            self.batters[0]: (simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Batter"]==self.batters[0])&(simteamstats.ballresult["Innings Type"]=="Batting")].value_counts(normalize=True,sort=False)*batadv).add(simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
            ,
            self.batters[1]: (simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Batter"]==self.batters[1])&(simteamstats.ballresult["Innings Type"]=="Batting")].value_counts(normalize=True,sort=False)*batadv).add(simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
            }

        if len(inningswicketfallP[self.batters[0]])<2 or sum(inningswicketfallP[self.batters[0]])<0.99:
            inningswicketfallP={
                self.batters[0]: (simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Batter"].isin(simteams[thisinnings]["battingorder"]))&(simteamstats.ballresult["Innings Type"]=="Batting")].value_counts(normalize=True,sort=False)*batadv).add(simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
                ,
                self.batters[1]: (simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Batter"].isin(simteams[thisinnings]["battingorder"]))&(simteamstats.ballresult["Innings Type"]=="Batting")].value_counts(normalize=True,sort=False)*batadv).add(simteamstats.ballresult['Out/NotOut'].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
                }

        return inningswicketfallP


    def scorefallbackplayersP(self,nthinnings,thisinnings, bowlingteam,simteams,simteamstats,thisover,hometeam):

        if hometeam==thisinnings:
            batadv=0.55
            bowladv=1-batadv
        if hometeam==bowlingteam:
            bowladv=0.55
            batadv=1-bowladv
        if hometeam==None:
            batadv=0.5
            bowladv=0.5

        inningsscorefallP = {
            self.batters[0]: (simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Batter"]==self.batters[0])&(simteamstats.ballresult["Innings Type"]=="Batting")].value_counts(normalize=True,sort=False)*batadv).add(simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
            ,
            self.batters[1]: (simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Batter"]==self.batters[1])&(simteamstats.ballresult["Innings Type"]=="Batting")].value_counts(normalize=True,sort=False)*batadv).add(simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
            }

        if len(inningsscorefallP[self.batters[0]])==0 or sum(inningsscorefallP[self.batters[0]])!=1:
            inningsscorefallP = {
                self.batters[0]: (simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Batter"].isin(simteams[thisinnings]["battingorder"]))&(simteamstats.ballresult["Innings Type"]=="Batting")].value_counts(normalize=True,sort=False)*batadv).add(simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
                ,
                self.batters[1]: (simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Batter"].isin(simteams[thisinnings]["battingorder"]))&(simteamstats.ballresult["Innings Type"]=="Batting")].value_counts(normalize=True,sort=False)*batadv).add(simteamstats.ballresult["Batter Score"].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))&(simteamstats.ballresult["Innings Type"]=="Bowling")].value_counts(normalize=True,sort=False)*bowladv,fill_value=0)
                }
        
        return inningsscorefallP

    def extrasfallbackplayersP(self,nthinnings,thisinnings, bowlingteam,simteams,simteamstats,thisover,hometeam):

        extrasP= simteamstats.ballresult["Extras"].loc[(simteamstats.ballresult["Bowler"].isin(simteams[bowlingteam]["bowlers"]))].value_counts(normalize=True,sort=False)

        return extrasP

    def over(self,rng,nthinnings,thisinnings, bowlingteam,statsmatchtype,thisover,simteams,simteamstats,hometeam, wicketfallP,scoreP,extrasP, fieldingextrasP):


        
        # ball outcome generator
        legaldeliveries = 0
        while legaldeliveries < 6:
            
            if len(wicketfallP[self.batters[0]])<2 or sum(wicketfallP[self.batters[0]])<0.99:
                    wicketfallP=playersim.wicketfallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)
            # Over based wicket rng
            wicket = rng.choice(wicketfallP[self.batters[0]].index, p=wicketfallP[self.batters[0]].tolist(), shuffle=False)
            

            if wicket=="Out":
                legaldeliveries += 1
                self.battersstats[self.batters[0]] += 1
                self.inningswickets += 1
                self.playersout.append(self.batters[0])

            if self.inningswickets == 10:
                self.inningsovers=float(f"{thisover}.{legaldeliveries}")
                break

            if wicket=="Out":
                self.batters[0]= simteams[thisinnings]["battingorder"][(self.inningswickets + 1)]
                self.battersstats[simteams[thisinnings]["battingorder"][(self.inningswickets + 1)]] = 0

                wicketfallP,scoreP = playersim.playerP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)

                if len(wicketfallP[self.batters[0]])<2 or sum(wicketfallP[self.batters[0]])<0.99:
                    wicketfallP=playersim.wicketfallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)

                if len(scoreP[self.batters[0]])==0 or sum(scoreP[self.batters[0]])<0.99:
                    scoreP=playersim.scorefallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)

                # Fix if p-values don't have all possibilites
                if len(scoreP[self.batters[0]])>0 and len(scoreP[self.batters[0]])<7:
                    scoreP = playersim.redistributepvalues(self,scoreP)

            # Over based scoring rng
            if wicket=="Not Out":

                if len(scoreP[self.batters[0]])==0 or sum(scoreP[self.batters[0]])<0.99:
                    scoreP=playersim.scorefallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)

                # Fix if p-values don't have all possibilites
                if len(scoreP[self.batters[0]])>0 and len(scoreP[self.batters[0]])<7:
                    scoreP = playersim.redistributepvalues(self,scoreP)

                #try:
                batterscore = rng.choice(scoreP[self.batters[0]].index, p=scoreP[self.batters[0]].tolist(), shuffle=False)
                # except:
                #     print(self.batters[0])

                if batterscore!=0:
                    extras = rng.choice(extrasP.index, p=extrasP.tolist(), shuffle=False)
                    self.inningsscore+=(batterscore+extras)
                    if extras==0:
                        legaldeliveries += 1
                        self.battersstats[self.batters[0]] += 1
                    if (batterscore)%2!=0:
                        self.batters.reverse()
                    if extras!=0:
                        self.battersstats[self.batters[0]] += 1
                if batterscore==0:
                    legaldeliveries += 1
                    self.battersstats[self.batters[0]] += 1
                    fieldingextras = rng.choice(fieldingextrasP.index, p=fieldingextrasP.tolist(), shuffle=False)
                    self.inningsscore += (fieldingextras)
                    if (fieldingextras)%2!=0:
                        self.batters.reverse()


            # Check if score has been chased
            if ((statsmatchtype == "T20" or statsmatchtype == "ODI" or statsmatchtype == "ODM") and (nthinnings == 1 and self.inningsscore > self.matchresults["Innings 1 Score"][-1])) or (statsmatchtype == "Test" and (self.inningswickets==10 or (nthinnings == 3 and self.inningsscore > (self.matchresults["Innings 1 Score"][-1]+self.matchresults["Innings 3 Score"][-1]-self.matchresults["Innings 2 Score"][-1])))):
                self.inningsovers=float(f"{thisover}.{legaldeliveries}")
                break
        

    def redistributepvalues(self,scoreP):
        # print(f"Before Modification: {scoreP}")
        lowestP = min(scoreP[self.batters[0]])
        scoreP[self.batters[0]]=scoreP[self.batters[0]] /(1+lowestP)
        newPs= (1-sum(scoreP[self.batters[0]]))/(7-len(scoreP[self.batters[0]]))
        # print(scoreP)
        scoreP[self.batters[0]]=scoreP[self.batters[0]].reindex([0,1,2,3,4,5,6],fill_value=newPs)
        # print(scoreP)
        # for eachvalue in scoreP[self.batters[0]].index:
        #     if pd.isna(scoreP[self.batters[0]][eachvalue]):
        #         scoreP[self.batters[0]][eachvalue]=newPs
        # print(scoreP)
        # print(f"ScoreP Sum: {sum(scoreP)}")
        return scoreP

    def resetinningstally(self):
        self.inningsscore = 0
        self.inningswickets = 0
        self.inningsovers = 0
        self.playersout=[]
        self.batters=[]
        self.battersstats={}

    def mcsimulations(self,statsmatchtype,simulations,inningsorder,rain,matchscore,hometeam):
        # print("Sims started")
        start = time.time()
        # setup random generator object
        rng = np.random.default_rng()
        # Set function dictionary
        # matchtypes={"T20": ld.limitedovers, "ODI": ld.limitedovers,"ODM": ld.limitedovers,"Test": tm.testmatch}
        classtypes={"T20": ld, "ODI": ld,"ODM": ld,"Test": tm}

        sim=classtypes[statsmatchtype]()

        sim.resultssetup()
        for thismatch in range(simulations):
            
            sim.matchresultssetup()
            # Function to simulate a match
            if statsmatchtype=="T20" or statsmatchtype=="ODI" or statsmatchtype=="ODM":
                sim.limitedovers(rng,statsmatchtype,inningsorder,rain,self.simteams,self.simteamstats,matchscore,hometeam,start)

            if statsmatchtype=="Test":
                sim.testmatch(rng,statsmatchtype,inningsorder,rain,self.simteams,self.simteamstats,matchscore,hometeam,start)

        return sim.results
            
    def sim(self, statsdatabase, statsfrom_date, statsto_date, statssex, statsmatchtype,simulations,inningsorder=None,rain=False,matchscore=None,hometeam=None, multicore=True):
        # Setup match results
        playersim.simresultssetup(self,statsmatchtype)

        # Search for pvalues
        playersim.pvaluesearch(self, statsdatabase, statsfrom_date, statsto_date, statssex, statsmatchtype)

        # multiprocessing
        if multicore==True:
            cores = os.cpu_count()
            procpool=mp.Pool(cores)

            simulations=int(simulations/cores)
            inputs=None
            inputs=[]
            print(f"Sims/cpu: {simulations}")
            for x in range(cores):
                inputs.append((self,statsmatchtype,simulations,inningsorder,rain,matchscore,hometeam))

            #start = time.time()
            simprocs = procpool.starmap(playersim.mcsimulations,inputs)
            
            procpool.close()
            #print(f'Time after mcsimulations(): {time.time() - start}')

            for eachdict in simprocs:
                for eachlist in eachdict:
                    self.simresults[eachlist].extend(eachdict[eachlist])
            print("Sims finished")

            self.simresults=pd.DataFrame(self.simresults)

        if multicore==False:
            start = time.time()
            rng = np.random.default_rng()
            # print(f'Time after rng: {time.time() - start}')

            # Set function dictionary
            classtypes={"T20": ld, "ODI": ld,"ODM": ld,"Test": tm}

            sim=classtypes[statsmatchtype]()
            # simulation generator
            sim.resultssetup()
            for thismatch in range(simulations):
                sim.matchresultssetup()
                # Function to simulate a match
                if statsmatchtype=="T20" or statsmatchtype=="ODI" or statsmatchtype=="ODM":
                    sim.limitedovers(rng,statsmatchtype,inningsorder,rain,self.simteams,self.simteamstats,matchscore,hometeam,start)

                if statsmatchtype=="Test":
                    sim.testmatch(rng,statsmatchtype,inningsorder,rain,self.simteams,self.simteamstats,matchscore,hometeam,start)

            self.simresults=pd.DataFrame(sim.results)



class ld(playersim):
    def __init__(self) -> None:
        self.inningswickets = 0
        self.inningsscore = 0
        self.inningsovers = 0
        self.matchresults = None
        self.results = None
        self.batters=[]
        self.battersstats={}
        self.playersout = []

    def resultssetup(self):
        self.results = {"Innings 1 Team":[], "Innings 1 Wickets":[],"Innings 1 Score":[], "Innings 1 Overs":[], "Innings 2 Team":[],"Innings 2 Wickets":[], "Innings 2 Score":[], "Innings 2 Overs":[],"Winner":[]}
    def matchresultssetup(self):
        self.matchresults = {"Innings 1 Team":[], "Innings 1 Wickets":[],"Innings 1 Score":[], "Innings 1 Overs":[], "Innings 2 Team":[],"Innings 2 Wickets":[], "Innings 2 Score":[], "Innings 2 Overs":[],"Winner":[]}

    def midinningssetup(self, matchscore):
        if len(matchscore) == 1:
            self.inningswickets = matchscore["Innings 1"][1]
            self.inningsscore = matchscore["Innings 1"][2]
            self.inningsovers = matchscore["Innings 1"][3]
        if len(matchscore) == 2:
            self.matchresults["Innings 1 Team"].append(matchscore["Innings 1"][0])
            self.matchresults["Innings 1 Wickets"].append(matchscore["Innings 1"][1])
            self.matchresults["Innings 1 Score"].append(matchscore["Innings 1"][2])
            self.matchresults["Innings 1 Overs"].append(matchscore["Innings 1"][3])
            self.inningswickets = matchscore["Innings 2"][1]
            self.inningsscore = matchscore["Innings 2"][2]
            self.inningsovers = matchscore["Innings 2"][3]

    def limitedovers(self,rng,statsmatchtype,inningsorder,rain,simteams,simteamstats,matchscore,hometeam,start):

        if matchscore:
            ld.midinningssetup(self,matchscore)

        # Randomly set innings order if not given
        if not inningsorder:
            # toss rng to decide inningsorder
            inningsorder = rng.choice(list(simteams.keys()), p=[0.5,0.5],size=2, replace=False, shuffle=False).tolist()

        # innings generator
        for nthinnings, thisinnings in enumerate(inningsorder):

            if matchscore and (len(matchscore)==2 and thisinnings == matchscore["Innings 1"][0]):
                continue
            # set bowling team
            bowlingteam = None
            for eachteam in inningsorder:
                if eachteam!=thisinnings:
                    bowlingteam = eachteam

            if statsmatchtype=="T20":
                overs=20
            if statsmatchtype=="ODI" or statsmatchtype=="ODM":
                overs=50

            # over generator
            for thisover in range(overs):
                # skip to last matchscore overs
                if matchscore and ((len(matchscore)==1 and thisover<(matchscore["Innings 1"][3])) or (len(matchscore)==2 and thisover<(matchscore["Innings 2"][3]))):
                    continue

                if self.inningswickets == 10 or (nthinnings == 1 and self.inningsscore > self.matchresults["Innings 1 Score"][-1]):
                    break
                
                self.batters.reverse()
                if not self.batters:
                    self.batters = simteams[thisinnings]["battingorder"][0:2]
                    for eachbatter in self.batters:
                        self.battersstats[eachbatter] = 0

                wicketfallP,scoreP = playersim.playerP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)
                
                # Fix if p-values don't have all possibilites
                if len(scoreP[self.batters[0]])>0 and len(scoreP[self.batters[0]])<7:
                    scoreP = playersim.redistributepvalues(self,scoreP)

                if len(wicketfallP[self.batters[0]])<2 or sum(wicketfallP[self.batters[0]])<0.99:
                    wicketfallP=playersim.wicketfallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)

                if len(scoreP[self.batters[0]])==0 or sum(scoreP[self.batters[0]])<0.99:
                    scoreP=playersim.scorefallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)

                extrasP, fieldingextrasP = ld.extras(self,nthinnings,thisinnings, bowlingteam,simteams,simteamstats,thisover)
                if len(extrasP)==0 or sum(extrasP)<0.99:
                    extrasP=playersim.extrasfallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)
                
                ld.over(self,rng,nthinnings,thisinnings, bowlingteam,statsmatchtype,thisover,simteams,simteamstats,hometeam,wicketfallP,scoreP,extrasP, fieldingextrasP)

                if self.inningswickets == 10 or (nthinnings == 1 and self.inningsscore > self.matchresults["Innings 1 Score"][-1]) or (self.inningswickets+1) == len(simteams[thisinnings]["battingorder"]):
                    break
                self.inningsovers+=1

            if self.inningsovers==0:
                self.inningsovers=overs
            self.matchresults[f"Innings {nthinnings+1} Team"].append(thisinnings)
            self.matchresults[f"Innings {nthinnings+1} Wickets"].append(self.inningswickets)
            self.matchresults[f"Innings {nthinnings+1} Score"].append(self.inningsscore)
            self.matchresults[f"Innings {nthinnings+1} Overs"].append(self.inningsovers)
            # print(self.battersstats)
            ld.resetinningstally(self)

        
        if self.matchresults["Innings 1 Score"][-1]>self.matchresults["Innings 2 Score"][-1]:
            self.matchresults["Winner"].append(self.matchresults["Innings 1 Team"][-1])
        if self.matchresults["Innings 1 Score"][-1]<self.matchresults["Innings 2 Score"][-1]:
            self.matchresults["Winner"].append(self.matchresults["Innings 2 Team"][-1])
        if self.matchresults["Innings 1 Score"][-1]==self.matchresults["Innings 2 Score"][-1]:
            self.matchresults["Winner"].append("Tie")
        for eachresult in self.results:
            self.results[eachresult].extend(self.matchresults[eachresult])

class tm(playersim):
    def __init__(self) -> None:
        self.inningswickets = 0
        self.inningsscore = 0
        self.inningsovers = 0
        self.matchresults = None
        self.results = None
        self.batters=[]
        self.battersstats={}
        self.playersout = []

    def resultssetup(self):
        self.results={"Innings 1 Team":[],"Innings 1 Wickets":[], "Innings 1 Score":[], "Innings 1 Overs":[],"Innings 2 Team":[],"Innings 2 Wickets":[], "Innings 2 Score":[],"Innings 2 Overs":[],"Innings 3 Team":[],"Innings 3 Wickets":[], "Innings 3 Score":[],"Innings 3 Overs":[],"Innings 4 Team":[], "Innings 4 Wickets":[],"Innings 4 Score":[],"Innings 4 Overs":[],"Winner":[]}

    def matchresultssetup(self):
        self.matchresults={"Innings 1 Team":[],"Innings 1 Wickets":[], "Innings 1 Score":[], "Innings 1 Overs":[],"Innings 2 Team":[],"Innings 2 Wickets":[], "Innings 2 Score":[],"Innings 2 Overs":[],"Innings 3 Team":[],"Innings 3 Wickets":[], "Innings 3 Score":[],"Innings 3 Overs":[],"Innings 4 Team":[], "Innings 4 Wickets":[],"Innings 4 Score":[],"Innings 4 Overs":[],"Winner":[]}

    def midinningssetup(self, matchscore):
        if len(matchscore) == 1:
            self.inningswickets = matchscore["Innings 1"][1]
            self.inningsscore = matchscore["Innings 1"][2]
            self.inningsovers = matchscore["Innings 1"][3]

        if len(matchscore) == 2:
            self.matchresults["Innings 1 Team"].append(matchscore["Innings 1"][0])
            self.matchresults["Innings 1 Wickets"].append(matchscore["Innings 1"][1])
            self.matchresults["Innings 1 Score"].append(matchscore["Innings 1"][2])
            self.matchresults["Innings 1 Overs"].append(matchscore["Innings 1"][3])
            self.inningswickets = matchscore["Innings 2"][1]
            self.inningsscore = matchscore["Innings 2"][2]
            self.inningsovers = matchscore["Innings 2"][3]

        if len(matchscore) == 3:
            self.matchresults["Innings 1 Team"].append(matchscore["Innings 1"][0])
            self.matchresults["Innings 1 Wickets"].append(matchscore["Innings 1"][1])
            self.matchresults["Innings 1 Score"].append(matchscore["Innings 1"][2])
            self.matchresults["Innings 1 Overs"].append(matchscore["Innings 1"][3])
            self.matchresults["Innings 2 Team"].append(matchscore["Innings 2"][0])
            self.matchresults["Innings 2 Wickets"].append(matchscore["Innings 2"][1])
            self.matchresults["Innings 2 Score"].append(matchscore["Innings 2"][2])
            self.matchresults["Innings 2 Overs"].append(matchscore["Innings 2"][3])
            self.inningswickets = matchscore["Innings 3"][1]
            self.inningsscore = matchscore["Innings 3"][2]
            self.inningsovers = matchscore["Innings 3"][3]

        if len(matchscore) == 4:
            self.matchresults["Innings 1 Team"].append(matchscore["Innings 1"][0])
            self.matchresults["Innings 1 Wickets"].append(matchscore["Innings 1"][1])
            self.matchresults["Innings 1 Score"].append(matchscore["Innings 1"][2])
            self.matchresults["Innings 1 Overs"].append(matchscore["Innings 1"][3])
            self.matchresults["Innings 2 Team"].append(matchscore["Innings 2"][0])
            self.matchresults["Innings 2 Wickets"].append(matchscore["Innings 2"][1])
            self.matchresults["Innings 2 Score"].append(matchscore["Innings 2"][2])
            self.matchresults["Innings 2 Overs"].append(matchscore["Innings 2"][3])
            self.matchresults["Innings 3 Team"].append(matchscore["Innings 3"][0])
            self.matchresults["Innings 3 Wickets"].append(matchscore["Innings 3"][1])
            self.matchresults["Innings 3 Score"].append(matchscore["Innings 3"][2])
            self.matchresults["Innings 3 Overs"].append(matchscore["Innings 3"][3])
            self.inningswickets = matchscore["Innings 4"][1]
            self.inningsscore = matchscore["Innings 4"][2]
            self.inningsovers = matchscore["Innings 4"][3]

    def testmatch(self,rng,statsmatchtype,inningsorder,rain,simteams,simteamstats,matchscore,hometeam,start):

        if matchscore:
            tm.midinningssetup(self,matchscore)
        
        # Randomly set innings if not given
        if not inningsorder:
            # toss rng to decide inningsorder
            toss = rng.choice(list(simteams.keys()), p=[0.5,0.5],size=2, replace=False, shuffle=False).tolist()
            inningsorder = toss+toss
        if len(inningsorder)==2:
            inningsorder = [inningsorder[0],inningsorder[1],inningsorder[0],inningsorder[1]]

        
        if rain:
            rainaffected = rng.choice(["rain","no_rain"], p=[0.1,0.9],size=1, replace=False, shuffle=False)
        
        matchover = 0
        # skip to last matchscore overs
        if matchscore:
            thisover = self.inningsovers
            for eachscore in matchscore:
                matchover += matchscore[eachscore][3]
        # inningsnumber=0
        # innings generator
        for nthinnings, thisinnings in enumerate(inningsorder):
            if matchscore and nthinnings<(len(matchscore)-1):
                continue

            # set bowling team
            bowlingteam = None
            for eachteam in inningsorder:
                if eachteam!=thisinnings:
                    bowlingteam = eachteam

            thisover=0
            # Innings overs generator
            
            while matchover<450:

                if (nthinnings == 3 and self.inningsscore > (self.matchresults["Innings 1 Score"][-1]+self.matchresults["Innings 3 Score"][-1]-self.matchresults["Innings 2 Score"][-1])) or (nthinnings==0 and (((self.inningsscore)/(450-matchover))>2)) or (nthinnings==1 and (((self.inningsscore-self.matchresults["Innings 1 Score"][-1])/(450-matchover))>2)) or (nthinnings==2 and (((self.inningsscore+self.matchresults["Innings 1 Score"][-1]-self.matchresults["Innings 2 Score"][-1])/(450-matchover))>4)):
                    break

                self.batters.reverse()
                if not self.batters:
                    self.batters = simteams[thisinnings]["battingorder"][0:2]
                    for eachbatter in self.batters:
                        self.battersstats[eachbatter] = 0

                wicketfallP,scoreP = playersim.playerP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)
                # Fix if p-values don't have all possibilites
                if len(scoreP[self.batters[0]])>0 and len(scoreP[self.batters[0]])<7:
                    scoreP = playersim.redistributepvalues(self,scoreP)

                if len(wicketfallP[self.batters[0]])<2 or sum(wicketfallP[self.batters[0]])<0.99:
                    wicketfallP=playersim.wicketfallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)

                if len(scoreP[self.batters[0]])==0 or sum(scoreP[self.batters[0]])<0.99:
                    scoreP=playersim.scorefallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam)

                extrasP, fieldingextrasP = ld.extras(self,nthinnings,thisinnings, bowlingteam,simteams,simteamstats,thisover)
                if len(extrasP)==0 or sum(extrasP)<0.99:
                    extrasP=playersim.extrasfallbackplayersP(self,nthinnings,thisinnings,bowlingteam,simteams,simteamstats,thisover,hometeam,)
                #print(f'Time after beforeover: {time.time() - start}')
                # Over generator
                tm.over(self,rng,nthinnings,thisinnings, bowlingteam,statsmatchtype,thisover,simteams,simteamstats,hometeam,wicketfallP,scoreP,extrasP, fieldingextrasP)

                # print(f'Time after afterover: {time.time() - start}')

                if self.inningswickets==10 or (nthinnings == 3 and self.inningsscore > (self.matchresults["Innings 1 Score"][-1]+self.matchresults["Innings 3 Score"][-1]-self.matchresults["Innings 2 Score"][-1])) or (nthinnings==2 and (((self.inningsscore+self.matchresults["Innings 1 Score"][-1]-self.matchresults["Innings 2 Score"][-1])/(450-matchover))>4)):
                    break

                thisover +=1
                matchover +=1

                # rain rng
                if rain:
                    if rainaffected=="rain":
                        rain=False
                        rainaffected="no_rain"
                        overslost = rng.integers(low=15, high=60)
                        matchover+=overslost
            
            #print(thisover)

            # Record innings results
            if self.inningsovers==0:
                self.inningsovers= thisover
            self.matchresults[f"Innings {nthinnings+1} Team"].append(thisinnings)
            self.matchresults[f"Innings {nthinnings+1} Wickets"].append(self.inningswickets)
            self.matchresults[f"Innings {nthinnings+1} Score"].append(self.inningsscore)
            self.matchresults[f"Innings {nthinnings+1} Overs"].append(self.inningsovers)
            tm.resetinningstally(self)
        
        # print(f"Overs left: {450-matchover}")

        # Record match winners
        #print(self.matchresults)
        if (self.matchresults["Innings 1 Score"][-1]+self.matchresults["Innings 3 Score"][-1])>(self.matchresults["Innings 2 Score"][-1]+self.matchresults["Innings 4 Score"][-1]) and (self.matchresults["Innings 4 Wickets"][-1]<10):
            self.matchresults["Winner"].append("Draw")
        if (self.matchresults["Innings 1 Score"][-1]+self.matchresults["Innings 3 Score"][-1])>(self.matchresults["Innings 2 Score"][-1]+self.matchresults["Innings 4 Score"][-1]) and (self.matchresults["Innings 4 Wickets"][-1]==10):
            self.matchresults["Winner"].append(self.matchresults["Innings 1 Team"][-1])
        if (self.matchresults["Innings 1 Score"][-1]+self.matchresults["Innings 3 Score"][-1])<(self.matchresults["Innings 2 Score"][-1]+self.matchresults["Innings 4 Score"][-1]):
            self.matchresults["Winner"].append(self.matchresults["Innings 2 Team"][-1])
        if (self.matchresults["Innings 1 Score"][-1]+self.matchresults["Innings 3 Score"][-1])==(self.matchresults["Innings 2 Score"][-1]+self.matchresults["Innings 4 Score"][-1]):
            self.matchresults["Winner"].append("Tie")

        # Record match results in return value
        for eachresult in self.results:
            self.results[eachresult].extend(self.matchresults[eachresult])



""" 
cricketstats is a script for getting team and player statistics from the cricsheet.org database for data analysis.
Copyright (C) 2021  Saranga Sudarshan

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>. 
"""

# This is an example script of how you can use cricketstats

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# If you installed cricketstats by pip import the module with line below
import cricketstats as cks

# If you cloned the github repo uncomment the block below.
"""
import os
import sys
module_path = os.path.abspath(os.path.join("./cricketstats/src/"))
if module_path not in sys.path:
    sys.path.append(module_path)
"""


""" 1. Create a search object """
# For ease of use first declare a list of players or teams to search stats for and then call that variable
Ozbatters = ["DA Warner", "SPD Smith",  "M Labuschagne",  "TM Head", "MS Harris", "C Green", "AT Carey", "UT Khawaja", "MR Marsh", "GJ Maxwell", "MP Stoinis", "MS Wade", "JP Inglis"]

bblteams = ["Sydney Thunder", "Melbourne Renegades", "Brisbane Heat", "Sydney Sixers", "Melbourne Stars", "Perth Scorchers", "Adelaide Strikers", "Hobart Hurricanes"]

# create a search object with "cricketstats.search" based on the inputs above. Input arguments: players=[] or teams=[] You can create multiple objects at the same time.
search1 = cks.cricketstats.search(players=Ozbatters)

# If you want to find stats for all players ot teams use the "allplayers=True" or "allteams=True" option in the class.
# search2 = cricketstats.search(allplayers=True)

# You can see below another search object setup for BBL teams. Just uncomment the line to use it.
# search3 = cricketstats.search(teams=bblteams)

""" 2. Apply the "stats()" method to the search object with the necessary arguments. """
# Arguments are mostly lists, with items separated by commas.

# First we can setup our arguments.
# Required input arguments:
database = "/home/saranga/Downloads/all_json.zip" # Path of cricsheet.org's all matches json zip file on your system. You can download the file at https://cricsheet.org/downloads/
from_date = (2018, 10, 1) # The (Year, Month, Day) from which to start search
to_date = (2021, 12, 31) # The (Year, Month, Day) from which to end search
matchtype = ["Test"] # Options: ["Test", "MDM", "ODI", "ODM", "T20", "IT20"] Explanation of what these mean is found at https://cricsheet.org/downloads/

# Optional Arguments:
# First we can set overs and innings related arguments:
betweenovers=[] # Search only these overs. eg. betweenovers = [1, 20]. Or if you only want to search for powerplays that are mandatory and option input "powerplays" eg. betweenovers =["powerplays"]
innings=[] # Search these innings. Options: 1, 2, 3, 4 eg. innings = [1,3]
fielders=[] # Search bowling stats involves these fielders.
oppositionbatters=[] # Search overs where players have bowled against certain batters. Options: batter names. You can also put in a list of batters by handedness. eg. oppositionbatters=["Left hand", "Right Hand"]
oppositionbowlers=[] # Search overs where players have batted against certain bowlers Options: bowler names You can also put in list of bowlers by type, eg. oppositionbowlers=["Right arm pace","Left arm pace","Right arm Off break","Right arm Leg break","Left arm orthodox","Left arm wrist spin"]
superover=None # Search normal innings or superover innings. Options: True, False eg. superover=True
battingposition=[] # Search stats at certain position in batting order. eg. battingposition=[1,2,3]
bowlingposition=[] # Search stats at certain position in bowling order. eg. bowlingposition=[1,2]

# Second we have match related arguments
sex=[] # Search only matches of certain sex. Options: "male", "female" eg. sex = ["female"] 
playerteams=[] # Search matches where players have played in certain teams. Options: team names eg. playerteams=["Australia", "England"]
teammates=[] # Search matches where certain teammates play. For this to work playerteams must be populated.
oppositionteams=[] # Search matches where opposition is only certain teams. Options: team names eg. oppositionteams=["India", "England"]
venue=[] # Search matches played only at these venues Options: Cricket Grounds eg. venue=["Sydney Cricket Ground", "Melbourne Cricket Ground", ""Brisbane Cricket Ground, Woolloongabba""]
teamtype=[] # Search only for particualr type of teams. eg. "international", "club".
event=[] # Search matches played as part of these Leagues or Tournaments Options: Name of League or Tournament eg. event=["Sheffield Shield", "ICC World Cup", "Big Bash League"] 
matchresult=None # Search matches where players or teams have these results. When looking at players, this option **must** be used with at one team in playersteams variable. Options either "won" or "loss" or "draw" or "tie" eg. matchresult="won"
sumstats=False # When switched to True, it adds an "all players" or "all teams" row at end of result that sums all players or teams stats that are searched for.

# Apply stats() method applied on search object. You have to apply the method to every search object if you want the script to actually do teh search.
search1.stats(database, from_date, to_date, matchtype, betweenovers=betweenovers, innings=innings, sex=sex, playersteams=playerteams, oppositionbatters=oppositionbatters, oppositionbowlers=oppositionbowlers, oppositionteams=oppositionteams, venue=venue, event=event, matchresult=matchresult, superover=superover, battingposition=battingposition, bowlingposition=bowlingposition, fielders=fielders, sumstats=sumstats)

# You can use the above template if you want or put all the values inside the brackets of the "stats()" method like I've done below for the search2 object.
# search2.stats(database="/home/saranga/Downloads/all_json.zip", from_date=(2018, 10, 1), to_date=(2021, 12, 31), matchtype=["T20"], betweenovers=[], innings=[], sex=[], playersteams=[], oppositionbatters=[], oppositionbowlers=[], oppositionteams=[], venue=[], event=["Big Bash League"], matchresult=[], superover=None, battingposition=[], bowlingposition=[], fielders=[], sumstats=False)


""" 3. Print result. Output is a pandas dataframe. """
# The output of the search is a pandas dataframe.
# print(search1.result.columns) # Use this line to print all the stats that are recorded and can be displayed.

# The line below prints the columns of the stats. There are way more stats than just the four used in the line below.
print(search1.result[["Games", "Batting Avg", "Score MeanAD", "Balls Faced"]]) 

# To get the long form data of each innings the players/team have played use the line below. This is best displayed as dataframe in jupyter notebook.
print(search1.inningsresult)

# If you want to save the search result as a csv file that you can open in excel you can use the line below.
# search1.result.to_csv(./YOUR FILE LOCATION)

""" 4. Plotting """
# You can use the plotting methods from the pandas package or matplotlib.pyplot for plotting.
# Handy tip, use the following list compression to extract all scores from multiple players or teams: [x for sublist in dataframe for x in sublist]

#search1.result.sort_values(by="Balls Faced", ascending=False)[["Balls Faced", "Batting Avg", "Score MeanAD"]].plot(kind="bar", rot=15, fontsize=8, title="Tests since 2018/2019 (data: cricsheet.org)", secondary_y = "Balls Faced")
# plt.legend()
# plt.show()


""" 5. Bonus Data Analysis Example """
# Pythagorean Expected Win Percentage for a team: This is a formula for calculating expected win percentage given runs scored and conceded by team
# Formula: Expected Win % = (Runs scored)^exponent / ((Runs scored)^exponent + (Runs conceded)^exponent)
# Exponent Options: Vine2016: 7.41, SenevirathneaManage2021: maximum likelihood method, 4.71 for ODI matches and 6.06 for Twenty20. The least squares method, 5.01 and 6.56 respectively for ODI and Twenty20.
#search2.result["Exp Win %"] = ((pow(search2.result["Runs"], 7.41) / (pow(search2.result["Runs"], 7.41) + pow(search2.result["Runsgiven"], 7.41)))*100)
#print(search2.result[["Win %", "Exp Win %"]])